from __future__ import annotations

import shutil
from enum import IntEnum
from pathlib import Path

from pydantic import (
    BaseModel,
    Field,
    FilePath,
    field_validator,
)

__all__ = [
    "HMMName",
    "DBName",
    "HMMFile",
    "DBFile",
    "NewDBFile",
    "SnapName",
    "SnapFile",
    "NewSnapFile",
    "Gencode",
    "NAME_MAX_LENGTH",
    "HMM_NAME_PATTERN",
    "DB_NAME_PATTERN",
    "SNAP_NAME_PATTERN",
]


def _file_name_pattern(ext: str):
    return r"^[0-9a-zA-Z_\-.][0-9a-zA-Z_\-. ]+\." + ext + "$"


NAME_MAX_LENGTH = 128

HMM_NAME_PATTERN = _file_name_pattern("hmm")
DB_NAME_PATTERN = _file_name_pattern("dcp")
SNAP_NAME_PATTERN = _file_name_pattern("dcs")


class HMMName(BaseModel):
    name: str = Field(pattern=HMM_NAME_PATTERN, max_length=NAME_MAX_LENGTH)

    @property
    def db_name(self):
        return DBName(name=self.name[:-4] + ".dcp")


class DBName(BaseModel):
    name: str = Field(pattern=DB_NAME_PATTERN, max_length=NAME_MAX_LENGTH)

    @property
    def hmm_file_name(self):
        return HMMName(name=self.name[:-4] + ".hmm")


class SnapName(BaseModel):
    name: str = Field(pattern=SNAP_NAME_PATTERN, max_length=NAME_MAX_LENGTH)


class HMMFile(BaseModel):
    path: FilePath

    @field_validator("path")
    def must_have_extension(cls, x: FilePath):
        if x.suffix != ".hmm":
            raise ValueError("must end in `.hmm`")
        return x

    @property
    def _dbpath(self) -> Path:
        return self.path.parent / f"{self.path.stem}.dcp"

    @property
    def dbfile(self) -> DBFile:
        return DBFile(path=self._dbpath)

    @property
    def newdbfile(self) -> NewDBFile:
        return NewDBFile(path=self._dbpath)


class DBFile(BaseModel):
    path: FilePath

    @field_validator("path")
    def must_have_extension(cls, x: FilePath):
        if x.suffix != ".dcp":
            raise ValueError("must end in `.dcp`")
        return x


class NewDBFile(BaseModel):
    path: Path

    @field_validator("path")
    def must_have_extension(cls, x: Path):
        if x.suffix != ".dcp":
            raise ValueError("must end in `.dcp`")
        return x

    @field_validator("path")
    def must_not_exist(cls, x: Path):
        if x.exists():
            raise ValueError("path already exists")
        return x


class SnapFile(BaseModel):
    path: FilePath

    @field_validator("path")
    def must_have_extension(cls, x: FilePath):
        if x.suffix != ".dcs":
            raise ValueError("must end in `.dcs`")
        return x


class NewSnapFile(BaseModel):
    path: Path

    @classmethod
    def create_from_prefix(cls, prefix: str):
        try:
            x = cls(path=Path(f"{prefix}.dcs").absolute())
        except ValueError:
            for i in range(1, 1001):
                try:
                    x = cls(path=Path(f"{prefix}.{i}.dcs").absolute())
                except ValueError:
                    continue
                else:
                    break
            else:
                raise ValueError(
                    f"failed to find a noncolliding filename for prefix {prefix}"
                )
        return x

    @field_validator("path")
    def must_have_extension(cls, x: Path):
        if x.suffix != ".dcs":
            raise ValueError("must end in `.dcs`")
        return x

    @field_validator("path")
    def must_not_exist(cls, x: Path):
        if x.exists():
            x.unlink()
        return x

    @field_validator("path")
    def basedir_must_not_exist(cls, x: Path):
        if basedir(x).exists():
            raise ValueError(f"`{basedir(x)}` path must not exist")
        return x

    @property
    def basename(self):
        return basedir(self.path)

    def make_archive(self):
        basename = self.basename
        x = shutil.make_archive(str(basename), "zip", self.path.parent, basename.name)
        shutil.move(x, self.path)
        shutil.rmtree(basename)


def basedir(x: Path):
    return x.parent / str(x.stem)


class Gencode(IntEnum):
    """NCBI genetic codes."""

    SGC0 = 1
    SGC1 = 2
    SGC2 = 3
    SGC3 = 4
    SGC4 = 5
    SGC5 = 6
    SGC8 = 9
    SGC9 = 10
    BAPP = 11
    AYN = 12
    AMC = 13
    AFMC = 14
    BMN = 15
    CMC = 16
    TMC = 21
    SOMC = 22
    TMMC = 23
    PMMC = 24
    CDSR1G = 25
    PTN = 26
    KN = 27
    CN = 28
    MN = 29
    PN = 30
    BN = 31
    BP = 32
    CMMC = 33

    # IntEnum of Python3.10 returns a different string representation.
    # Make it return the same as in Python3.11
    def __str__(self):
        return str(self.value)



from deciphon_core.cffi import ffi, lib

__all__ = ["DeciphonError"]


class DeciphonError(RuntimeError):
    def __init__(self, errno: int):
        msg = ffi.string(lib.error_string(errno)).decode()
        super().__init__(msg)



import os
import re
import shutil
import sys
import sysconfig
import tarfile
import urllib.request
from dataclasses import dataclass
from pathlib import Path
from subprocess import check_call

RPATH = "$ORIGIN" if sys.platform.startswith("linux") else "@loader_path"

PWD = Path(os.path.dirname(os.path.abspath(__file__)))
TMP = PWD / ".build_ext"
PKG = PWD / "deciphon_core"
INTERFACE = PKG / "interface.h"

BIN = Path(PKG) / "bin"
LIB = Path(PKG) / "lib"
INCL = Path(PKG) / "include"
EXTRA = f"-Wl,-rpath,{RPATH}/lib"
SHARE = Path(PKG) / "share"

CMAKE_OPTS = [
    "-DCMAKE_BUILD_TYPE=Release",
    "-DBUILD_SHARED_LIBS=ON",
    f"-DCMAKE_INSTALL_RPATH={RPATH}",
]

CPM_OPTS = ["-DCPM_USE_LOCAL_PACKAGES=ON"]

NNG_OPTS = [
    "-DCMAKE_INSTALL_LIBDIR=lib",
    "-DNNG_TESTS=OFF",
    "-DNNG_TOOLS=OFF",
    "-DNNG_ENABLE_NNGCAT=OFF",
]


@dataclass
class Ext:
    github_user: str
    github_project: str
    git_tag: str
    root_dir: str
    cmake_opts: list[str]


EXTS = [
    Ext("EBI-Metagenomics", "lip", "v0.5.4", "./", CMAKE_OPTS),
    Ext(
        "EBI-Metagenomics",
        "hmmer3",
        "hmmer-reader-v0.7.4",
        "./hmmer-reader",
        CMAKE_OPTS,
    ),
    Ext("EBI-Metagenomics", "imm", "v5.5.0", "./", CMAKE_OPTS + CPM_OPTS),
    Ext("nanomsg", "nng", "v1.5.2", "./", CMAKE_OPTS + NNG_OPTS),
    Ext(
        "EBI-Metagenomics",
        "hmmer3",
        "h3client-v0.12.6",
        "./h3client",
        CMAKE_OPTS + CPM_OPTS,
    ),
    Ext(
        "EBI-Metagenomics",
        "deciphon",
        "c-core-v0.20.7",
        "./c-core",
        CMAKE_OPTS + CPM_OPTS,
    ),
]


def rm(folder: Path, pattern: str):
    for filename in folder.glob(pattern):
        filename.unlink()


def resolve_bin(bin: str):
    paths = [sysconfig.get_path("scripts", x) for x in sysconfig.get_scheme_names()]
    paths += os.path.expandvars(os.getenv("PATH", "")).split(os.pathsep)
    for x in paths:
        y = Path(x) / bin
        if y.exists():
            return str(y)
    raise RuntimeError(f"Failed to find {bin}.")


def build_ext(ext: Ext):
    from cmake import CMAKE_BIN_DIR

    url = (
        f"https://github.com/{ext.github_user}/{ext.github_project}"
        f"/archive/refs/tags/{ext.git_tag}.tar.gz"
    )
    tar_filename = f"{ext.github_project}-{ext.git_tag}.tar.gz"

    os.makedirs(TMP, exist_ok=True)
    with open(TMP / tar_filename, "wb") as lf:
        lf.write(urllib.request.urlopen(url).read())

    with tarfile.open(TMP / tar_filename) as tf:
        dir = os.path.commonprefix(tf.getnames())
        tf.extractall(TMP)

    prj_dir = TMP / dir / ext.root_dir
    bld_dir = prj_dir / "build"
    os.makedirs(bld_dir, exist_ok=True)

    cmake = [str(v) for v in Path(CMAKE_BIN_DIR).glob("cmake*")][0]
    check_call([cmake, "-S", str(prj_dir), "-B", str(bld_dir)] + ext.cmake_opts)
    n = os.cpu_count()
    check_call([cmake, "--build", str(bld_dir), "-j", str(n), "--config", "Release"])

    check_call([cmake, "--install", str(bld_dir), "--prefix", str(PKG)])


if __name__ == "__main__":
    from cffi import FFI

    ffibuilder = FFI()

    rm(PKG, "cffi.*")
    rm(PKG / "lib", "**/lib*")
    shutil.rmtree(TMP, ignore_errors=True)

    if not os.environ.get("DECIPHON_CORE_DEVELOP", False):
        if sys.platform == "linux":
            patch = [resolve_bin("patchelf"), "--set-rpath", "$ORIGIN"]
            for lib in LIB.glob("*.so*"):
                check_call(patch + [str(lib)])

    if not os.environ.get("DECIPHON_CORE_DEVELOP", False):
        for ext in EXTS:
            build_ext(ext)

    libs = os.environ.get("DECIPHON_CORE_LIB_PATH", "").split(";")
    incls = os.environ.get("DECIPHON_CORE_INCLUDE_PATH", "").split(";")

    libs = [x for x in libs if len(x) > 0]
    incls = [x for x in incls if len(x) > 0]

    ffibuilder.cdef(open(INTERFACE, "r").read())
    ffibuilder.set_source(
        "deciphon_core.cffi",
        """
        #include "deciphon/deciphon.h"
        #include "h3client/h3client.h"
        """,
        language="c",
        libraries=["deciphon", "h3client"],
        library_dirs=libs + [str(LIB)],
        include_dirs=incls + [str(INCL)],
        extra_link_args=[str(EXTRA)],
    )
    ffibuilder.compile(verbose=True)

    shutil.rmtree(BIN, ignore_errors=True)
    shutil.rmtree(INCL, ignore_errors=True)
    shutil.rmtree(SHARE, ignore_errors=True)
    shutil.rmtree(LIB / "cmake", ignore_errors=True)

    if not os.environ.get("DECIPHON_CORE_DEVELOP", False):
        if sys.platform == "linux":
            patch = [resolve_bin("patchelf"), "--set-rpath", "$ORIGIN"]
            for lib in LIB.glob("*.so*"):
                check_call(patch + [str(lib)])

        find = ["/usr/bin/find", str(LIB), "-type", "l"]
        exec0 = ["-exec", "/bin/cp", "{}", "{}.tmp", ";"]
        exec1 = ["-exec", "/bin/mv", "{}.tmp", "{}", ";"]
        check_call(find + exec0 + exec1)

        for x in list(LIB.iterdir()):
            linux_pattern = r"lib[^.]*\.so\.[0-9]+"
            macos_pattern = r"lib[^.]*\.[0-9]+\.dylib"
            pattern = r"^(" + linux_pattern + r"|" + macos_pattern + r")$"
            if not re.match(pattern, x.name):
                x.unlink()



from __future__ import annotations


from deciphon_core.cffi import ffi, lib
from deciphon_core.error import DeciphonError
from deciphon_core.params import Params
from deciphon_core.schema import DBFile, NewSnapFile
from deciphon_core.sequence import Sequence

__all__ = ["Scan"]


def check_exception(*_):
    return True


@ffi.def_extern(onerror=check_exception)
def interrupt(_):
    return False


class Scan:
    def __init__(self, params: Params, dbfile: DBFile):
        self._cscan = lib.scan_new(params.cparams)
        if self._cscan == ffi.NULL:
            raise MemoryError()
        self._dbfile = dbfile

    def dial(self, port: int = 51371):
        if rc := lib.scan_dial(self._cscan, port):
            raise DeciphonError(rc)

    def open(self):
        if rc := lib.scan_open(self._cscan, bytes(self._dbfile.path)):
            raise DeciphonError(rc)

    def close(self):
        if rc := lib.scan_close(self._cscan):
            raise DeciphonError(rc)

    def add(self, sequence: Sequence):
        id = sequence.id
        name = sequence.name.encode()
        data = sequence.data.encode()
        if rc := lib.scan_add(self._cscan, id, name, data):
            raise DeciphonError(rc)

    def run(self, snap: NewSnapFile):
        if rc := lib.scan_run(self._cscan, str(snap.basename).encode(), lib.interrupt, ffi.NULL):
            raise DeciphonError(rc)

    def interrupted(self) -> bool:
        return lib.scan_interrupted(self._cscan)

    def progress(self) -> int:
        return lib.scan_progress(self._cscan)

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *_):
        self.close()

    def __del__(self):
        if getattr(self, "_cscan", ffi.NULL) != ffi.NULL:
            lib.scan_del(self._cscan)



from deciphon_core.cffi import ffi, lib
from deciphon_core.error import DeciphonError

__all__ = ["Params"]


class Params:
    def __init__(self, num_threads: int, multi_hits: bool, hmmer3_compat: bool):
        self._cptr = ffi.new("struct params *")
        if self._cptr == ffi.NULL:
            raise MemoryError()

        if rc := lib.params_setup(self._cptr, num_threads, multi_hits, hmmer3_compat):
            raise DeciphonError(rc)

    @property
    def cparams(self):
        return self._cptr[0]

    def __del__(self):
        if getattr(self, "_cparams", ffi.NULL) != ffi.NULL:
            lib.press_del(self._cptr)
